/* vim: set ts=2 sw=2 tw=99 noet: */

function AddI16(lhs, rhs)
{
	var t = lhs + rhs;
	if ((lhs ^ rhs) & 0x8000) {
		if (t & 0x8000)
			t |= 0xFFFF0000;
		else
			t &= 0x0000FFFF;
	}
	return t;
}

function AndI16(lhs, rhs)
{
	var t = lhs & rhs;
	if (t & 0x8000)
		t |= 0xFFFF0000;
	else
		t &= 0x0000FFFF;
	return t;
}

function NotI16(v)
{
	v = ~v;
	if (v & 0x8000)
		v |= 0xFFFF0000;
	else
		v &= 0x0000FFFF;
	return v;
}

function Context()
{
	this.regs  = new Array(8);
	this.mem   = new Array(65536);
	this.pc    = 0x3000;
	this.psr   = 0;

	for (var i = 0; i < this.regs.length; i++)
		this.regs[i] = 0;
	for (var i = 0; i < this.mem.length; i++)
		this.mem[i] = 0;

	/*** Initialize TRAP routines ***/

	/* HALT */
	this.mem[0x25]  = 0x200;
	this.mem[0x200] = 0x2002; /* LD R0, PC+2 <= halt */
	this.mem[0x201] = 0x2202; /* LD R1, PC+2 <= mcr */
	this.mem[0x202] = 0x7040; /* STR [R1], R0 */
	this.mem[0x203] = 0x8000; /* bit 15 = halt */
	this.mem[0x204] = 0xFFFE; /* addr of MCR */

	/*** Set up dummy main program. */
	this.mem[0x3000] = 0xF025; /* HALT */

	this.setCC = function (t)
	{
		var flags = (t < 0) << 2;
		flags |= (t == 0) << 1;
		flags |= (t > 0);
		this.psr &= 0xFFFC;
		this.psr |= flags;
	}

	this.printState = function()
	{
		var s = '\t';
		for (var i = 0; i < this.regs.length; i++)
			s += 'R' + i + '=x' + this.regs[i].toString(16) + ' ';
		print('\tPSR = ' + this.psr.toString(16));
		print(s);
	}

	this.loadImage = function(image)
	{
		this.pc = image.start;
		for (var i = 0; i < image.data.length; i++)
			this.mem[this.pc + i] = image.data[i];
		this.regs[6] = this.pc;
	}

	this.getFlags = function()
	{
		return this.psr & 0x3;
	}

	this.isPrivileged = function ()
	{
		return this.psr & (1 << 15);
	}

	this.readU16 = function (addr)
	{
		addr &= 0x0000FFFF;
		return this.mem[addr] & 0x0000FFFF;
	}

	this.readI16 = function (addr)
	{
		addr &= 0x0000FFFF;
		return this.mem[addr];
	}

	this.writeI16 = function (addr, val)
	{
		addr &= 0x0000FFFF;
		this.mem[addr] = val;
	}

	this.execute = function ()
	{
		var br=0, dr=0, sr=0, sr1=0, sr2=0, imm=0, flags=0;
		var op=0, instruction=0;

		for (;;) {
			if (cx.readI16(0xFFFE) & (1 << 15))
				break;
			instruction = cx.readU16(cx.pc);
			//print(cx.pc.toString(16) + ': ' + instruction.toString(16));
			//cx.printState();
			cx.pc = AddI16(cx.pc, 1);
			op = instruction >> 12;
			switch (op)
			{
				/* BR */
				case 0:
					flags = (instruction >> 9);
					if (flags & cx.getFlags()) {
						imm = instruction & 0x1FF;
						if (imm & 0x100)
							imm |= 0xFFFFFE00;
						cx.pc = AddI16(cx.pc, imm);
					}
					break;

				/* ADD */
				case 1:
					dr = (instruction >> 9) & 0x7;
					sr1 = (instruction >> 6) & 0x7;
					if (!(instruction & (1 << 5))) {
						if (instruction & 0x38)
							throw "Invalid ADD instruction";
						sr2 = instruction & 0x7;
						imm = AddI16(cx.regs[sr1], cx.regs[sr2]);
					} else {
						imm = instruction & 0x1F;
						if (imm & 0x10)
							imm |= 0xFFFFFE0;
						imm = AddI16(cx.regs[sr1], imm);
					}
					cx.regs[dr] = imm;
					cx.setCC(imm);
					break;

				/* LD */
				case 2:
					dr = (instruction >> 9) & 0x7;
					imm = instruction & 0x1FF;
					if (imm & 0x100)
						imm |= 0xFFFFFE00;
					imm = AddI16(cx.pc, imm);
					imm = cx.readI16(imm);
					cx.regs[dr] = imm;
					cx.setCC(imm);
					break;

				/* ST */
				case 3:
					sr = (instruction >> 9) & 0x7;
					imm = instruction & 0x1FF;
					if (imm & 0x100)
						imm |= 0xFFFFFE00;
					imm = AddI16(cx.pc, imm);
					cx.writeI16(imm, cx.regs[sr]);
					break;

				/* JSR[R] */
				case 4:
					dr = cx.pc;
					if (instruction & (1 << 11)) {
						imm = instruction & 0x7FF;
						if (imm & 0x700)
							imm |= 0xFFFFF000;
						cx.pc = AddI16(cx.pc, imm);
					} else {
						if (instruction & 0xE3F)
							throw "Invalid JSRR instruction: " + instruction;
						sr = (instruction >> 6) & 0x7;
						cx.pc = cx.regs[sr];
					}
					cx.regs[7] = dr;
					break;

				/* AND */
				case 5:
					dr = (instruction >> 9) & 0x7;
					sr1 = (instruction >> 6) & 0x7;
					if (!(instruction & (1 << 5))) {
						if (instruction & 0x38)
							throw "Invalid AND instruction";
						sr2 = instruction & 0x7;
						imm = AndI16(cx.regs[sr1] + cx.regs[sr2]);
					} else {
						imm = instruction & 0x1F;
						if (imm & 0x10)
							imm |= 0xFFFFFE0;
						imm = AndI16(cx.regs[sr1], imm);
					}
					cx.regs[dr] = imm;
					cx.setCC(imm);
					break;

				/* LDR */
				case 6:
					dr = (instruction >> 9) & 0x7;
					sr = (instruction >> 6) & 0x7;
					imm = instruction & 0x3F;
					if (imm & 0x30)
						imm |= 0xFFFFFFC0;
					imm = AddI16(cx.regs[sr], imm);
					imm = cx.readI16(imm);
					cx.regs[dr] = imm;
					cx.setCC(imm);
					break;

				/* STR */
				case 7:
					sr = (instruction >> 9) & 0x7;
					br = (instruction >> 6) & 0x7;
					imm = instruction & 0x3F;
					if (imm & 0x30)
						imm |= 0xFFFFFFC0;
					imm = AddI16(cx.regs[br], imm);
					cx.writeI16(imm, cx.regs[sr]);
					break;

				/* RTI */
				case 8:
					if (cx.isPrivileged()) {
						cx.pc = cx.readI16(cx.regs[6]);
						cx.regs[6] = AddI16(cx.regs[6], 1);
						imm = cx.readI16(cx.regs[6]);
						cx.regs[6] = AddI16(cx.regs[6], 1);
						cx.psr = imm;
					} else {
						throw 1;
					}
					break;

				/* NOT */
				case 9:
					if ((instruction & 0x3F) != 0x3F)
						throw "Invalid NOT instruction: " + instruction;
					dr = (instruction >> 9) & 0x7;
					sr = (instruction >> 6) & 0x7;
					imm = NotI16(cx.regs[sr]);
					cx.regs[dr] = imm;
					cx.setCC(imm);
					break;

				/* LDI */
				case 10:
					dr = (instruction >> 9) & 0x7;
					imm = instruction & 0x1FF;
					if (imm & 0x100)
						imm |= 0xFFFFFE00;
					imm = AddI16(cx.pc, imm);
					imm = cx.readI16(imm);
					imm = cx.readI16(imm);
					cx.regs[dr] = imm;
					cx.setCC(imm);
					break;

				/* STI */
				case 11:
					sr = (instruction >> 9) & 0x7;
					imm = instruction & 0x100;
					if (imm & 0x100)
						imm |= 0xFFFFFE00;
					imm = AddI16(cx.pc, imm);
					imm = cx.readI16(imm);
					cx.writeI16(imm, cx.regs[sr]);
					break;

				/* JMP, RET */
				case 12:
					if (instruction & 0xE3F)
						throw "Invalid JMP instruction: " + instruction;
					sr = (instruction >> 6) & 0x7;
					cx.pc = cx.regs[sr];
					break;

				/* ILLEGAL */
				case 13:
					throw 2;

				/* LEA */
				case 14:
					dr = (instruction >> 9) & 0x7;
					imm = instruction & 0x1FF;
					if (imm & 0x100)
						imm |= 0xFFFFFE00;
					imm = AddI16(cx.pc, imm);
					cx.regs[dr] = imm;
					cx.setCC(imm);
					break;

				/* TRAP */
				case 15:
					cx.regs[7] = cx.pc;
					imm = instruction & 0xFF;
					cx.pc = cx.readI16(imm);
					break;
			}
		}
	}
}
